﻿using System;
using System.IO;
using System.Text;

namespace Parade.Metadata.Id3
{
	/// <summary>
	/// ID3v1、ID3v2.2 ～ 4 のタグ情報をダンプします。
	/// </summary>
	class TagDumper : IDisposable
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="filePath">ファイルのパス情報。</param>
		/// <param name="dumpFilePath">ダンプ先となるテキストファイルへのパス情報。</param>
		public TagDumper( string filePath, string dumpFilePath )
		{
			this._reader   = new FileStream( filePath, FileMode.Open, FileAccess.Read, FileShare.Read );
			this._writer   = new StreamWriter( dumpFilePath, true, Encoding.Unicode );
			this._filePath = filePath;
		}
	
		/// <summary>
		/// リソースを解放します。
		/// </summary>
		public void Dispose()
		{
			if( this._reader != null )
			{
				this._reader.Dispose();
				this._reader = null;
			}

			if( this._writer != null )
			{
				this._writer.Dispose();
				this._writer = null;
			}

			GC.SuppressFinalize( this );
		}

		/// <summary>
		/// フレームをダンプします。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="version">バージョン情報。</param>
		/// <param name="tagSize">タグ情報のサイズ。</param>
		private void DumpFramesV2( Stream reader, StreamWriter writer, byte[] version, uint tagSize )
		{
			writer.WriteLine( "\nFrames\nType\tValue" );

			Func< Stream, StreamWriter, uint, int, uint > dumpFrame = null;
			if( version[ 0 ] == 2 )
			{
				dumpFrame = new Func< Stream, StreamWriter, uint, int, uint >( DumpFrameV22 );
			}
			else if( version[ 0 ] == 3 && version[ 1 ] == 0 )
			{
				dumpFrame = new Func< Stream, StreamWriter, uint, int, uint >( DumpFrameV23 );
			}
			else
			{
				dumpFrame = new Func< Stream, StreamWriter, uint, int, uint >( DumpFrameV24 );
			}

			var remainSize = tagSize;
			var index      = 0;
			while( remainSize > 0 )
			{
				remainSize -= dumpFrame( reader, writer, remainSize, index++ );
			}
		}

		/// <summary>
		/// フレームをダンプします。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="info">タグ情報。</param>
		/// <param name="size">タグ情報のサイズ。</param>
		/// <param name="index">フレームの通し番号。</param>
		private void DumpFrameValue( Stream reader, StreamWriter writer, Id3TagInfo info, uint size, int index )
		{
			if( info.Type == Id3TagDataType.Text )
			{
				// エンコード指定は 1 バイトなので、文字列を読み取るサイズ指定は必ず 1 バイト分、差し引く必要がある
				var encode = ( Id3TextEncoding )reader.ReadByte();
				var str    = ReadEncodeString( reader, encode, size - 1 );
				writer.WriteLine( "[ {0} ] Value\t{1} : {2}", index, encode, str );
			}
			else if( info.Type == Id3TagDataType.TextSimple )
			{
				// URL などはシステム標準エンコードとみなす
				var str = reader.ReadStringDefault( ( int )size );
				writer.WriteLine( "[ {0} ] Value\t{1}", index, str );
			}
			else
			{
				// 未対応のデータはとりあえずバイナリにしておく
				writer.WriteLine( "[ {0} ] Value\t byte[ {1} ]", index, size );
				reader.Seek( size, SeekOrigin.Current );
			}
		}

		/// <summary>
		/// ID3v2.2 のフレームをダンプします。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="remainSize">タグ情報全体の残りサイズ。</param>
		/// <param name="index">フレームの通し番号。</param>
		/// <returns>読み取ったデータ サイズ。</returns>
		private uint DumpFrameV22( Stream reader, StreamWriter writer, uint remainSize, int index )
		{
			// フレームヘッダがないので何もしない
			if( remainSize < 6 ) { return 0; }

			var id = reader.ReadStringAscii( 3 );
			writer.WriteLine( "[ {0} ] ID\t{1}", index, id );

			uint size = 0;
			{
				var b = reader.ReadBytes( 3 );
				size = ( ( uint )( b[ 0 ] << 16 ) | ( uint )( b[ 1 ] << 8 ) | ( uint )( b[ 2 ] ) ); 
				writer.WriteLine( "[ {0} ] Size\t{1}", index, size );
			}

			// 未定義フレームなら終了
			var info = Id3DefinedTags.V22.GetInfo( id );
			if( info == null )
			{
				writer.WriteLine( "[ {0} ] Value\tUndefined frame", index );
				return size + 6;
			}

			this.DumpFrameValue( reader, writer, info, size, index );
			return size + 6;
		}

		/// <summary>
		/// ID3v2.3 のフレームをダンプします。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="remainSize">タグ情報全体の残りサイズ。</param>
		/// <param name="index">フレームの通し番号。</param>
		/// <returns>読み取ったデータ サイズ。</returns>
		private uint DumpFrameV23( Stream reader, StreamWriter writer, uint remainSize, int index )
		{
			// フレームヘッダがないので何もしない
			if( remainSize < 10 ) { return 0; }

			var id = reader.ReadStringAscii( 4 );
			writer.WriteLine( "[ {0} ] ID\t{1}", index, id );

			uint size = 0;
			{
				var b = reader.ReadBytes( 4 );
				size = ( ( uint )( b[ 0 ] << 24 ) | ( uint )( b[ 1 ] << 16 ) | ( uint )( b[ 2 ] << 8 ) | ( uint )( b[ 3 ] ) );
				writer.WriteLine( "[ {0} ] Size\t{1}", index, size );
			}

			// 未定義フレームなら終了
			var info = Id3DefinedTags.V23.GetInfo( id );
			if( info == null )
			{
				writer.WriteLine( "[ {0} ] Value\tUndefined frame", index );
				return size + 10;
			}

			// フラグ
			uint optionalSize = 0;
			{
				// ステータス
				var b = reader.ReadBytes( 1 );
				var status = Convert.ToString( b[ 0 ], 2 );

				// フォーマット
				b = reader.ReadBytes( 1 );
				var format = Convert.ToString( b[ 0 ], 2 );

				// 圧縮ありなら、伸長後のサイズ ( 4 ) をスキップ
				if( ( b[ 0 ] & 0x80 ) != 0 ) { optionalSize += 4; }

				// 暗号化ありなら、その方法を示すフラグ ( 1 ) をスキップ
				if( ( b[ 0 ] & 0x40 ) != 0 ) { ++optionalSize; }

				// グループ化ありなら、その識別子 ( 1 ) をスキップ
				if( ( b[ 0 ] & 0x20 ) != 0 ) { ++optionalSize; }

				if( optionalSize > 0 )
				{
					reader.Seek( optionalSize, SeekOrigin.Current );
				}

				writer.WriteLine( "[ {0} ] Status, Format flags\t{1}, {2}", index, status, format );
			}
	
			this.DumpFrameValue( reader, writer, info, size, index );
			return size + 10 + optionalSize;
		}

		/// <summary>
		/// ID3v2.4 のフレームをダンプします。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="remainSize">タグ情報全体の残りサイズ。</param>
		/// <param name="index">フレームの通し番号。</param>
		/// <returns>読み取ったデータ サイズ。</returns>
		private uint DumpFrameV24( Stream reader, StreamWriter writer, uint remainSize, int index )
		{
			// フレームヘッダがないので何もしない
			if( remainSize < 10 ) { return 0; }

			var id = reader.ReadStringAscii( 4 );
			writer.WriteLine( "[ {0} ] ID\t{1}", index, id );

			uint size = 0;
			{
				var b = reader.ReadBytes( 4 );
				size = ( ( uint )( b[ 0 ] << 24 ) | ( uint )( b[ 1 ] << 16 ) | ( uint )( b[ 2 ] << 8 ) | ( uint )( b[ 3 ] ) );
				writer.WriteLine( "[ {0} ] Size\t{1}", index, size );
			}

			// 未定義フレームなら終了
			var info = Id3DefinedTags.V24.GetInfo( id );
			if( info == null )
			{
				writer.WriteLine( "[ {0} ] Value\tUndefined frame", index );
				return size + 10;
			}

			// フラグ
			uint optionalSize = 0;
			{
				// ステータス
				var b = reader.ReadBytes( 1 );
				var status = Convert.ToString( b[ 0 ], 2 );

				// フォーマット
				b = reader.ReadBytes( 1 );
				var format = Convert.ToString( b[ 0 ], 2 );

				// グループ化ありなら、その識別子 ( 1 ) をスキップ
				if( ( b[ 0 ] & 0x40 ) != 0 ) { ++optionalSize; }

				// 圧縮ありなら、伸長後のサイズ ( 4 ) をスキップ
				if( ( b[ 0 ] & 0x8 ) != 0 ) { optionalSize += 4; }

				// 暗号化ありなら、その方法を示すフラグ ( 1 ) をスキップ
				if( ( b[ 0 ] & 0x4 ) != 0 ) { ++optionalSize; }

				if( optionalSize > 0 )
				{
					reader.Seek( optionalSize, SeekOrigin.Current );
				}

				writer.WriteLine( "[ {0} ] Status, Format flags\t{1}, {2}", index, status, format );
			}

			this.DumpFrameValue( reader, writer, info, size, index );
			return size + 10 + optionalSize;
		}
		
		/// <summary>
		/// 拡張ヘッダをダンプし、新しいタグ情報のサイズを算出します。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="version">バージョン情報。</param>
		/// <param name="headerFlags">ヘッダのフラグ。</param>
		/// <param name="tagSize">タグ情報のサイズ。</param>
		/// <returns>新しいタグ情報のサイズ。</returns>
		private uint DumpExtenedeHeaderV2( Stream reader, StreamWriter writer, byte[] version, byte headerFlags, uint tagSize )
		{
			// 2.3 以降のみ
			if( version[ 0 ] < 3 || ( ( headerFlags & 0x40 ) == 0 ) ) { return tagSize; }

			writer.WriteLine( "\nExtended Header\nType\tValue" );

			// 拡張ヘッダのサイズとデータ部分を差し引く
			var exSize = this.DumpSizeV2( reader, writer, "Extended header size" );
			tagSize -= exSize + 4;

			// サイズ以降のデータ構成はバージョンによって異なる
			if( version[ 0 ] == 3 && version[ 1 ] == 0 )
			{
				var exFlags = reader.ReadBytes( 2 );
				writer.WriteLine( "Flags\t{0:X2}, {1:X2}", exFlags[ 0 ], exFlags[ 1 ] );
				this.DumpSizeV2( reader, writer, "Padding size" );

				if( ( exFlags[ 0 ] & 0x80 ) != 0 )
				{
					var crc = reader.ReadBytes( 4 );
					writer.WriteLine( "CRC\t{0:X2}, {1:X2}, {2:X2}, {3:X2}", crc[ 0 ], crc[ 1 ], crc[ 2 ], crc[ 3 ] );
				}
			}
			else
			{
				var flagLength = reader.ReadBytes( 1 );
				writer.WriteLine( "Flag length\t{0}", flagLength[ 0 ] );

				var exFlags        = reader.ReadBytes( 1 );
				var hasCrc         = ( ( exFlags[ 0 ] & 0x20 ) != 0 );
				var isRestrictions = ( ( exFlags[ 0 ] & 0x10 ) != 0 );
				writer.WriteLine( "Flags\t{0:X2} : Tag updated = {1}, CRC = {2}, Tag restrictions = {3}", exFlags[ 0 ], ( ( exFlags[ 0 ] & 0x40 ) != 0 ), hasCrc, isRestrictions );

				var remainSize = exSize - 4;

				if( hasCrc && remainSize >= 5 )
				{
					var crc = reader.ReadBytes( 5 );
					writer.WriteLine( "CRC\t{0:X2}, {1:X2}, {2:X2}, {3:X2}, {4:X2}", crc[ 0 ], crc[ 1 ], crc[ 2 ], crc[ 3 ], crc[ 4 ] );
					remainSize -= 5;
				}

				if( isRestrictions && remainSize >= 1 )
				{
					var b = reader.ReadBytes( 1 );
					writer.WriteLine( "Tag restrictions\t{0}", Convert.ToString( b[ 0 ], 2 ) );
					--remainSize;
				}
			}

			return tagSize;
		}

		/// <summary>
		/// ヘッダのフラグをダンプします。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <returns>フラグ。</returns>
		private byte DumpHeaderFlagsV2( Stream reader, StreamWriter writer )
		{
			// 8 ビットの値について「非同期化」、「拡張ヘッダ所有」、「実験タグ」の順にビットフラグが設けられている
			var b = reader.ReadBytes( 1 );
			writer.WriteLine( "Flags\t{0} : Unsync = {1}, Extended Header = {2}, Experimental = {3}", b[ 0 ], ( ( b[ 0 ] & 0x80 ) != 0 ), ( ( b[ 0 ] & 0x40 ) != 0 ), ( ( b[ 0 ] & 0x20 ) != 0 ) );

			return b[ 0 ];
		}

		/// <summary>
		/// ID3v2 のタグ情報サイズを取得します。
		/// サイズは合計 4 バイトのデータで、第 7 ビットが 0 になるように設定されいるので実質的なデータは 3 バイト ( 28 ビット ) となります。
		/// 例えば 257 バイトの場合は { 00, 00, 02, 01 } のようにレイアウトされます。
		/// サイズにはヘッダ部分を含みません。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="type">サイズの種類。</param>
		/// <returns>サイズ。</returns>
		private uint DumpSizeV2( Stream reader, StreamWriter writer, string type )
		{
			var b    = reader.ReadBytes( 4 );
			var size = ( ( uint )( ( b[ 0 ] & 0x7f ) << 21 ) | ( uint )( ( b[ 1 ] & 0x7f ) << 14 ) | ( uint )( ( b[ 2 ] & 0x7f ) << 7 ) | ( uint )( b[ 0 ] & 0x7f ) );

			writer.WriteLine( "{0}\t{1}", type, size );
			return size;
		}

		/// <summary>
		/// バージョン情報をダンプします。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <returns>バージョン情報。</returns>
		/// <exception cref="NotSupportedException">サポート外のファイル形式です。</exception>
		private byte[] DumpVersionV2( Stream reader, StreamWriter writer )
		{
			// v2.2.0、v2.3.0、v2.3.1 ( v2.4 )、v2.4.0 のみサポート対象とする
			var b = reader.ReadBytes( 2 );
			if( !( ( b[ 0 ] == 2 && b[ 1 ] == 0 ) || ( b[ 0 ] == 3 && ( b[ 1 ] == 0 ) || b[ 1 ] == 1 ) || ( b[ 0 ] == 4 && b[ 1 ] == 0 ) ) )
			{
				throw new NotSupportedException( String.Format( "Unsupported version \"{0}.{1}\".", b[ 0 ], b[ 1 ] ) );
			}

			writer.WriteLine( "Version\tv2.{0}.{1}", b[ 0 ], b[ 1 ] );
			return b;
		}

		/// <summary>
		/// ID3v1 のタグ情報をダンプします。
		/// </summary>
		/// <exception cref="NotSupportedException">サポート外のファイル形式です。</exception>
		public void DumpV1()
		{
			var reader = this._reader;
			var writer = this._writer;

			// ID3v1 チェック
			reader.Seek( -128, SeekOrigin.End );
			if( !this.HasTagV1( reader ) ) { throw new NotSupportedException( "Id3v1 tag was not found." ); }
			
			writer.WriteLine( "\nID3v1\nType\tValue" );

			writer.WriteLine( "Title\t{0}",  reader.ReadStringDefault( 30 ) );
			writer.WriteLine( "Artist\t{0}", reader.ReadStringDefault( 30 ) );
			writer.WriteLine( "Album\t{0}",  reader.ReadStringDefault( 30 ) );
			writer.WriteLine( "Year\t{0}",   reader.ReadStringAscii( 4 )    );

			// コメントとトラック番号
			{
				int		track = 0;
				string	comment;

				var b = reader.ReadBytes( 30 );
				if( b[ 28 ] == 0 )
				{
					// 余白はゼロで埋められるため、29 バイト目がゼロならば以降はゼロまたはトラック番号と考えられる。
					// 仮にトラック番号が未指定でも、ゼロとして取得される。
					// 通常、トラック番号は 1 から始まるので、ゼロであれば無効値としてもよい。
					//
					track = b[ 29 ];
					comment = StringUtility.TrimLastNull( Encoding.Default.GetString( b, 0, 28 ) );
				}
				else
				{
					comment = StringUtility.TrimLastNull( Encoding.Default.GetString( b ) );
				}

				writer.WriteLine( "Comment\t{0}",      comment );
				writer.WriteLine( "Track Number\t{0}", track   );
			}

			writer.WriteLine( "Genre\t{0}", reader.ReadByte() );
		}

		/// <summary>
		/// ID3v2 のタグ情報をダンプします。
		/// </summary>
		/// <exception cref="NotSupportedException">サポート外のファイル形式です。</exception>
		public void DumpV2()
		{
			var reader = this._reader;
			var writer = this._writer;

			// ID3v2 タグ ( 3 )
			reader.Seek( 0, SeekOrigin.Begin );
			if( !this.HasTagV2( reader ) ) { throw new NotSupportedException( "Id3v2 tag was not found." ); }
			writer.WriteLine( "\nID3v2\n\nHeader\nType\tValue" );

			// バージョン ( 2 )
			byte[] version = this.DumpVersionV2( reader, writer );

			// フラグ ( 1 )
			var flags = this.DumpHeaderFlagsV2( reader, writer );

			// サイズ ( 4 )
			var size = this.DumpSizeV2( reader, writer, "Tag size" );

			// 非同期化されているならば、解除後のデータを生成して読み取りストリームを変更
			if( ( flags & 0x80 ) != 0 )
			{
				var m  = new MemoryStream();
				size   = this.Unsync( reader, m, size );
				reader = m;
			}

			size = this.DumpExtenedeHeaderV2( reader, writer, version, flags, size );

			// フレーム出力
			this.DumpFramesV2( reader, writer, version, size );
		}

		/// <summary>
		/// ID3v1 のタグ情報を持つことを調べます。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <returns>タグ情報を持つ場合は true。それ以外は false。</returns>
		private bool HasTagV1( Stream reader )
		{
			var b = reader.ReadBytes( 3 );
			return ( b[ 0 ] == TagDumper.Id3v1Tag[ 0 ] && b[ 1 ] == TagDumper.Id3v1Tag[ 1 ] && b[ 2 ] == TagDumper.Id3v1Tag[ 2 ] );
		}

		/// <summary>
		/// ID3v2 のタグ情報を持つことを調べます。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <returns>タグ情報を持つ場合は true。それ以外は false。</returns>
		private bool HasTagV2( Stream reader )
		{
			var b = reader.ReadBytes( 3 );
			return ( b[ 0 ] == TagDumper.Id3v2Tag[ 0 ] && b[ 1 ] == TagDumper.Id3v2Tag[ 1 ] && b[ 2 ] == TagDumper.Id3v2Tag[ 2 ] );
		}

		/// <summary>
		/// 文字コードを指定して文字列を読み取ります。
		/// </summary>
		/// <param name="reader">タグ情報を読み取るためのストリーム。</param>
		/// <param name="encode">文字コード。</param>
		/// <param name="size">データ サイズ。</param>
		/// <returns>文字列。</returns>
		private string ReadEncodeString( Stream reader, Id3TextEncoding encode, uint size )
		{
			switch( encode )
			{
			case Id3TextEncoding.Utf16:   return StringUtility.GetStringUtf16WithBom( reader.ReadBytes( ( int )size ) );
			case Id3TextEncoding.Utf16BE: return reader.ReadStringUtf16BE( ( int )size );
			case Id3TextEncoding.Utf8:    return reader.ReadStringUtf8( ( int )size );
			default:                      return reader.ReadStringDefault( ( int )size );
			}
		}
	
		/// <summary>
		/// 非同期化を解除します。
		/// </summary>
		/// <param name="src">非同期化を解除すべきデータを持つストリーム。</param>
		/// <param name="dest">非同期化を解除したデータを書き込むストリーム。</param>
		/// <param name="size">非同期化された状態のデータ サイズ。</param>
		/// <returns>非同期化を解除した後のデータ サイズ。</returns>
		private uint Unsync( Stream src, Stream dest, uint size )
		{
			var reader = new BinaryReader( src );
			var writer = new BinaryWriter( dest );
			var last     = ( byte )0;
			var count    = ( uint )0;
			var syncSize = ( uint )0;
			
			while( count < size )
			{
				var b = reader.ReadByte();
				if( last == 0xFF && b == 0x00 )
				{
					++syncSize;
				}
				else
				{
					writer.Write( b );
				}

				last = b;
				++count;
			}

			if( last == 0xFF )
			{
				writer.Write( ( byte)0x00 );
				++syncSize;
			}

			return syncSize;
		}

		/// <summary>
		/// ファイルのパス情報。
		/// </summary>
		private string _filePath;

		/// <summary>
		/// タグ情報を読み取るためのストリーム。
		/// </summary>
		private Stream _reader;

		/// <summary>
		/// タグ情報を出力するためのストリーム。
		/// </summary>
		private StreamWriter _writer;

		/// <summary>
		/// ID3v1 のタグ始点をあらわすバイト配列。ASCII 文字列で表すと "TAG" となる。
		/// </summary>
		private static readonly byte[] Id3v1Tag = { 0x54, 0x41, 0x47 };

		/// <summary>
		/// ID3v2 のタグ始点をあらわすバイト配列。ASCII 文字列で表すと "ID3" となる。
		/// </summary>
		private static readonly byte[] Id3v2Tag = { 0x49, 0x44, 0x33 };
	}
}
